
import re



sizeDict = {
    "void": 0,
    "char": 1,
    "unsigned char": 1,
    "signed char": 1,
    "short": 2,
    "unsigned short": 2,
    "int": 4,
    "unsigned int": 4,
    "long": 4,
    "unsigned long": 4,
    "long long": 8,
    "unsigned long long": 8,
    "float": 4,
    "double": 8,
    "long double": 16,  # This can vary, but 16 bytes is a common size
    "int8": 1,
    "uint8": 1,
    "int16": 2,
    "uint16": 2,
    "int32": 4,
    "uint32": 4,
    "int64": 8,
    "uint64": 8,
    "va_list": 4,
}

class FunctionArgument():
    def __init__(self, type : str, name:str):

        type = type.strip().replace("_t","")
        name = name.strip()

        self.name = name
        self.type = type
        self.size = 4

        if type in sizeDict:
            self.size = sizeDict[type]
        

class Function():
    def __init__(self, name, address = "???", args= [], type = 'void'):
        self.name = name
        self.address = address
        self.args = args 
        self.type = type
        self.locals = []
        self.calls = []
        self.raw = ""

    def addArgument(self, arg):
        self.args.append(arg)

    @classmethod
    def fromString(cls, line):
        
        #0x0040100 <int main(int, char**)>
        #int main(int argc, char **argv)
        #void doSomething(int a, int b)

        functionAddress = None

        regex = re.compile(r"(0x[0-9A-F]+)\S*?<(.*?)>")
        match = regex.match(line)

        if match:
            functionAddress = int(match.group(1), 16)
            line = match.group(2)
        

        #Parse the function definition
        regex = re.compile(r"(.*?[*]?)(\w+)\((.*?)\)")
        match = regex.match(line)

        if match:
            returnType = match.group(1).replace(" ","")
            functionName = match.group(2)
            args = match.group(3).split(',')

            functionArgs = []
            for arg in args:

                arg = arg.strip()

                if arg and arg != "...":
                    argParts = arg.split(' ')

                    if 'const' in argParts:
                        argParts.remove('const')

                    argType = argParts[0]
                    argName = None

                    if len(argParts) > 1:
                        argName = argParts[1]

                        while argName[0] == '*':
                            argType += '*'
                            argName = argName[1:]

                    functionArgs.append(FunctionArgument(argType, argName))
            
            ret = cls(functionName, functionAddress, functionArgs, returnType)
        
            return ret 

        else : 
            
            print(f"No match: {line}")

    def toYaml(self):
        return {
            'name': self.name,
            'address': self.address,
            'type': self.type,
            'args': [{'name': arg.name, 'type': arg.type} for arg in self.args],
            'locals': self.locals,
            'calls': self.calls
        }



def parseCFile(code):
    functions = []
    

    # Regular expressions to match function signatures, local variable declarations, and function calls
    function_re = re.compile(r'\b([A-Za-z_]\w*\s+\*?\s*\b[A-Za-z_]\w*\s*\([^)]*\)\s*)\{(.*?)}',re.DOTALL)

    call_re = re.compile(r'\b([A-Za-z_]\w*\s*\([^)]*\)\s*)')

    local_re = re.compile(r'(\w[\w\s\*]*\w)\s+(\w+)(\[(.*)\])?')

    defines_re = re.compile(r'#define\s+(\w+)\s+(\w+)')
    defines = {}

    define_matches = defines_re.finditer(code)

    for match in define_matches:
        defines[match.group(1)] = match.group(2)

        
    function_signatures = function_re.finditer(code)

    for match in function_signatures:
        signature = match.group(1).strip()
        function_body = match.group(2).strip()

        function = Function.fromString(signature)

        function.raw = match.group(0)


        
        if function:

            locals = local_re.finditer(function_body)


            for local in locals:
                
                if local.group(4) and local.group(4) in defines:
                    function.locals.append(f"{local.group(1)} {local.group(2)}[{defines[local.group(4)]}]")
                else:
                    function.locals.append(local.group(0))

            calls = call_re.finditer(function_body)

            for call in calls:
                function.calls.append(call.group(0))

            # function.locals = [var.strip() for var in local_re.findall(function_body)]
            # function.calls = [call.strip() for call in call_re.findall(function_body)]
            functions.append(function)

    return functions



import sys
import os

from staq.stackSession import StackSession

def test_session():

    session = StackSession()

    session.parseCommand("int i =42")

    wordCount = len(session.stack.currentFrame.cells[0].words)
    wordValue = session.stack.currentFrame.cells[0].words[0].value

    assert wordCount == 1
    assert wordValue == 42

def test_html():
    session = StackSession()
    session.stack.baseAddress =0xffffff


    session.parseCommand('int etc[64]')
    session.parseCommand('call main(1,"Hello World")')
    session.parseCommand('char buf[64]')
    session.parseCommand('call strcpy(buf,argv[0])')
    session.parseCommand('note 0xfeb3: note on local variable')


    html = session.stack.toHtml()

    with open("test.html", "w") as fh:
        fh.write(html)

def test_png():
    session = StackSession()
    session.stack.baseAddress =0xffffff
    imagePath = 'test.png'
    
    session.parseCommand('int etc[64]')
    session.parseCommand('call main(1,"Hello World")')
    session.parseCommand('char buf[64]')
    session.parseCommand('call strcpy(buf,argv[0])')
    session.parseCommand('note 0xfeb3: note on local variable')


    session.stack.generatePng(imagePath)

    assert os.path.exists(imagePath)


if __name__ == "__main__":
    test_session()



from staq.stack import Stack, StackFrame, StackCell 
from staq.function import Function, FunctionArgument
from staq.helper import *
from termcolor import colored

import re

class Register():
    def __init__(self, name, size, description):
        self.name = name
        self.size = size
        self.type = type
        self.value = None
        self.description = None


class CallingConvention():
    """
        Base Class for calling conventions
    """
    def __init__(self, name, session = None, stack = None):
        self.name = name
        self.endian = 'little'
        self.registers = {}
        self.session = session
        self.stack : Stack = stack

    def clear(self):

        for key in self.registers:
            self.registers[key].value = None


    def setInstructionPointer(self, val):
        pass

    def pop(self):
       return self.stack.pop()

    def push(self,cell):
        self.stack.push(cell)
    
    def popFrame(self):
       return self.stack.popFrame()
    
    def pushFrame(self, frame):
        self.stack.pushFrame(frame)

    def getReg(self, key):
        if key in self.registers:
            return self.registers[key].value
        
        return None
    
    def setReg(self,key, value):

        if key in self.registers:
            self.registers[key].value = value

    def leave(self):

        if self.stack.currentFrame:
            while self.stack.currentFrame.length() > 1:
                self.stack.pop()
            
            while self.stack.currentFrame.cells[0].size > 1:
                self.stack.pop()


    def ret(self):
        word = self.stack.pop()


        if word and word.value:
            self.jmp( word.value)
        else:
            self.jmp('???')

    def markArgs(self): 

        self.stack.argMarks = {}

        if self.stack.currentFrame and self.stack.currentFrame.function:
            func = self.stack.currentFrame.function

            pointer = self.stack.currentFrame.basePointer
            pointer+= 1 #skip the return address
            for i in range(len(func.args)):
                
                self.stack.argMarks[pointer] = func.args[i].name
                pointer += func.args[i].size



    def jmp(self, val):
        self.setInstructionPointer(val)

        offset = 0

        if val.startswith('<') and val.endswith('>'):
            val = val[1:-1]

        val = val.strip()

        if '+' in val:
            
            val, offset = val.split('+')


        if self.stack.currentFrame and self.stack.currentFrame.function and self.stack.currentFrame.function.name == val:
            pass 
        else:
            lastWord = self.stack.pop()
            func = self.session.functions.get(val, Function(val))
            self.stack.pushFrame(StackFrame(val,func))
            cell = StackCell('ret', [lastWord])

            self.stack.push(cell)

        self.markArgs()

    def updateRegs(self):
        self.setReg('esp', self.stack.pointer)


    def registersToAnsi(self, width = 30, color='yellow'):

        #  ┌───────────────────────────────────┐
        #  │ esp: <value>                      │
        #  │ ebp: <value>                      │
        #  │ eax: <value>                      │
        #  │ ebx: <value>                      │
        #  └───────────────────────────────────┘

        self.updateRegs()

        #get the max length of the register names
        maxNameLength = 0
        for key in self.registers:
            if len(key) > maxNameLength:
                maxNameLength = len(key)
        
        maxValLength =  width - (maxNameLength + 5)

        out = colored(f"┌{'─' * maxNameLength}{'─' * (maxValLength +5)}┐\n", color)
        for key in self.registers:
            if self.registers[key].value:

                if isinstance(self.registers[key].value, int):
                    strVal = hex(self.registers[key].value)
                else:
                    strVal = str(self.registers[key].value)

                truncVal = strVal.ljust(maxValLength)[:maxValLength]
                out += colored(f"│ {key.rjust(maxNameLength)} : {truncVal} │\n", color)

        out += colored(f"└{'─' * maxNameLength}{'─' * (maxValLength + 5)}┘\n", color)

        return out


    def call(self, callLine):


        regex = re.compile(r"(\S.*?)\((.*?)\)")

        match = regex.match(callLine)

        func : Function = None

        if match:

            functionName = match.group(1)

            args = match.group(2)

        else:
            functionName = callLine
            args = ""

        if args.strip() == "":
            args = []
        else:
            argStr = match.group(2)
            args = splitWithEscapes(argStr,',')


        # If the function does not exist, create it and assume all args are ints
        if functionName not in self.session.functions:
            newFunction = Function(functionName)

            for i in range(len(args)):
                newFunction.addArgument(FunctionArgument('int',f'arg{i+1}'))

            self.session.functions[functionName] = newFunction

        func  = self.session.functions[functionName]
        
        for i in reversed(range(len(args))):
            label = f'arg{i+1}'
            if i < len(func.args) and func.args[i].name: 
                label = f"{func.args[i].name}"
            

            cell = StackCell(  label, [args[i].strip()])
            self.stack.push(cell)

        currentFunction = self.session.getCurrentFunction()

        
        retCell = StackCell('ret', ['<???>'])
        if currentFunction:

            if isinstance(currentFunction, Function):
                
                retCell.setWords(f'<{currentFunction.name}+??>')
            else:
                retCell.setWords(f'<{currentFunction}+??>')

        self.stack.push(retCell)

        self.jmp(functionName)

        if func:
            for local in func.locals:
                cell = self.session.tryParseLocalVar(local)
                if cell:
                    self.stack.push(cell)


        

    


import sys
import os

from staq.helper import splitWithEscapes

def test_split():

   tp1 = 'hello {world: "hello world"}'
   tp2 = 'cmd: call printf(\"Hello: %d %d\", 1, 2)'

   split1 = splitWithEscapes(tp1, ":")
   split2 = splitWithEscapes(tp2, ":")

   assert split1 == [tp1]
   assert split2 == ['cmd', 'call printf(\"Hello: %d %d\", 1, 2)']

if __name__ == "__main__":
    test_split()